/*
 * talkatu
 * Copyright (C) 2017-2018 Gary Kramlich <grim@reaperworld.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
#include <gtk/gtk.h>

#include <talkatu/talkatubuffer.h>
#include <talkatu/talkatueditor.h>
#include <talkatu/talkatutoolbar.h>
#include <talkatu/talkatuview.h>

/**
 * SECTION:talkatueditor
 * @Title: Editor
 * @Short_description: A simple editor composed of other Talkatu Widgets
 *
 * TalkatuEditor is a composite widget of #TalkatuToolbar and #TalkatuView to
 * help make them easier to use.
 */

typedef struct {
	GtkWidget *view;
	GtkWidget *toolbar;
} TalkatuEditorPrivate;

G_DEFINE_TYPE_WITH_PRIVATE(TalkatuEditor, talkatu_editor, GTK_TYPE_BOX)

/******************************************************************************
 * Helpers
 *****************************************************************************/
static void
talkatu_editor_buffer_changed(TalkatuEditor *editor) {
	GActionGroup *action_group = NULL;
	TalkatuEditorPrivate *priv = talkatu_editor_get_instance_private(editor);
	GtkTextBuffer *buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(priv->view));

	if(TALKATU_IS_BUFFER(buffer)) {
		action_group = G_ACTION_GROUP(talkatu_buffer_get_action_group(TALKATU_BUFFER(buffer)));
	}

	gtk_widget_insert_action_group(priv->toolbar, "talkatu", action_group);
}

/******************************************************************************
 * Callbacks
 *****************************************************************************/
static void
talkatu_editor_view_buffer_changed_handler(GObject *obj, GParamSpec *pspec, gpointer data) {
	talkatu_editor_buffer_changed(TALKATU_EDITOR(data));
}

/******************************************************************************
 * GObject Stuff
 *****************************************************************************/
static void
talkatu_editor_init(TalkatuEditor *editor) {
	gtk_widget_init_template(GTK_WIDGET(editor));
}

static void
talkatu_editor_class_init(TalkatuEditorClass *klass) {
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(klass);

	/* load our template */
	gtk_widget_class_set_template_from_resource(
		GTK_WIDGET_CLASS(klass),
		"/org/bitbucket/pidgin/talkatu/ui/editor.ui"
	);

	gtk_widget_class_bind_template_child_internal_private(widget_class, TalkatuEditor, view);
	gtk_widget_class_bind_template_child_internal_private(widget_class, TalkatuEditor, toolbar);

	gtk_widget_class_bind_template_callback(widget_class, talkatu_editor_view_buffer_changed_handler);
}

/******************************************************************************
 * Public API
 *****************************************************************************/

/**
 * talkatu_editor_new:
 *
 * Creates a new #TalkatuEditor with a normal #GtkTextBuffer.
 *
 * Returns: (transfer full): The new #TalkatuEditor instance.
 */
GtkWidget *
talkatu_editor_new(void) {
	return GTK_WIDGET(g_object_new(TALKATU_TYPE_EDITOR, NULL));
}

/**
 * talkatu_editor_get_view:
 * @editor: The #TalkatuEditor instance.
 *
 * Gets the #TalkatuView that @editor is using.
 *
 * Returns: (transfer none): The #TalkatuView that @editor is using.
 */
GtkWidget *
talkatu_editor_get_view(TalkatuEditor *editor) {
	TalkatuEditorPrivate *priv = NULL;

	g_return_val_if_fail(TALKATU_IS_EDITOR(editor), NULL);

	priv = talkatu_editor_get_instance_private(editor);

	return priv->view;
}

/**
 * talkatu_editor_get_toolbar:
 * @editor: The #TalkatuEditor instance.
 *
 * Gets the #TalkatuToolbar that @editor is using.
 *
 * Returns: (transfer none): The #TalkatuToolbar that @editor is using.
 */
GtkWidget *
talkatu_editor_get_toolbar(TalkatuEditor *editor) {
	TalkatuEditorPrivate *priv = NULL;

	g_return_val_if_fail(TALKATU_IS_EDITOR(editor), NULL);

	priv = talkatu_editor_get_instance_private(editor);

	return priv->toolbar;
}
