/*
 * talkatu
 * Copyright (C) 2017-2018 Gary Kramlich <grim@reaperworld.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
#include <glib/gi18n-lib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gspell/gspell.h>
#include <gumbo.h>
#include <stdio.h>

#include "talkatu/talkatuactiongroup.h"
#include "talkatu/talkatubuffer.h"
#include "talkatu/talkatuenums.h"
#include "talkatu/talkatutag.h"
#include "talkatu/talkatuview.h"

/**
 * SECTION:talkatuview
 * @Title: Text View
 * @Short_description: Widget that displays a TalkatuBuffer
 *
 * #TalkatuView handles the display of a #TalkatuBuffer, the keybindings for
 * applying formats, as well as displaying links.
 */

/**
 * TalkatuViewSendBinding:
 * @TALKATU_VIEW_SEND_BINDING_RETURN: Represents the return.
 * @TALKATU_VIEW_SEND_BINDING_KP_ENTER: Represents the enter.
 * @TALKATU_VIEW_SEND_BINDING_SHIFT_RETURN: Represents shift-return.
 * @TALKATU_VIEW_SEND_BINDING_CONTROL_RETURN: Represents control-return.
 *
 * Flags for assigning and determining which key bindings should be used to
 * send a message.
 */

typedef struct {
	GSimpleActionGroup *action_group;
	TalkatuViewSendBinding send_binding;
	GspellTextView *gspell_view;

	/* this mark is used to keep track of our context for the context menu.  It
	 * is updated via cursor-moved and button-press callbacks.
	 */
	GtkTextMark *context_mark;
} TalkatuViewPrivate;

typedef struct {
	TalkatuView *view;
	const gchar *url;
} TalkatuViewOpenLinkData;

enum {
	PROP_0 = 0,
	PROP_SEND_BINDING,
	N_PROPERTIES,
};
static GParamSpec *properties[N_PROPERTIES];

enum {
	SIG_FORMAT_ACTIVATE,
	SIG_SHOULD_SEND_MESSAGE,
	SIG_SEND_MESSAGE,
	SIG_OPEN_URL,
	LAST_SIGNAL,
};
static guint signals[LAST_SIGNAL] = {0, };

G_DEFINE_TYPE_WITH_PRIVATE(TalkatuView, talkatu_view, GTK_TYPE_TEXT_VIEW)

/******************************************************************************
 * Callbacks
 *****************************************************************************/
static void
talkatu_view_insert_anchor_cb(GtkTextBuffer *buffer,
                              GtkTextIter *iter,
                              GtkTextChildAnchor *anchor,
                              gpointer user_data)
{
	GtkTextView *view = GTK_TEXT_VIEW(user_data);
	gpointer data = NULL;

	/* first check if talkatu-url is added */
	data = g_object_get_data(G_OBJECT(anchor), "talkatu-url");
	if(data != NULL) {
		GtkWidget *link = NULL;
		gpointer label = g_object_get_data(G_OBJECT(anchor), "talkatu-url-label");

		if(label != NULL) {
			link = gtk_link_button_new_with_label((gchar *)data, (gchar *)label);
		} else {
			link = gtk_link_button_new((gchar *)data);
		}

		gtk_text_view_add_child_at_anchor(view, link, anchor);

		gtk_widget_show(link);
	}
}

static void
talkatu_view_buffer_set_cb(GObject *view, GParamSpec *pspec, gpointer data) {
	TalkatuViewPrivate *priv = talkatu_view_get_instance_private(TALKATU_VIEW(view));
	GtkTextBuffer *buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
	GtkTextIter start;

	if(TALKATU_IS_BUFFER(buffer)) {
		priv->action_group = talkatu_buffer_get_action_group(TALKATU_BUFFER(buffer));
	}

	gtk_text_buffer_get_start_iter(buffer, &start);
	priv->context_mark = gtk_text_buffer_create_mark(buffer, NULL, &start, TRUE);

	gspell_text_view_basic_setup(priv->gspell_view);

	g_signal_connect_after(
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(view)),
		"insert-child-anchor",
		G_CALLBACK(talkatu_view_insert_anchor_cb),
		view
	);
}

/******************************************************************************
 * Default Signal Handlers
 *****************************************************************************/
static void
talkatu_view_format_activate(TalkatuView *view, const gchar *action_name) {
	TalkatuViewPrivate *priv = talkatu_view_get_instance_private(view);

	if(priv->action_group) {
		GAction *action = g_action_map_lookup_action(G_ACTION_MAP(priv->action_group), action_name);

		if(action) {
			g_action_activate(action, NULL);
		}
	}
}

static void
talkatu_view_should_send_message(TalkatuView *view, TalkatuViewSendBinding binding) {
	TalkatuViewPrivate *priv = talkatu_view_get_instance_private(view);

	if((priv->send_binding & binding) != 0) {
		g_signal_emit(view, signals[SIG_SEND_MESSAGE], 0);
	} else {
		GtkTextBuffer *buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));

		gtk_text_buffer_insert_at_cursor(buffer, "\n", 1);
	}
}

static gboolean
talkatu_view_popup_menu(GtkWidget *widget) {
	TalkatuViewPrivate *priv = talkatu_view_get_instance_private(TALKATU_VIEW(widget));
	GtkTextBuffer *buffer = NULL;
	GtkTextIter iter;

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget));

	gtk_text_buffer_get_iter_at_mark(
		buffer,
		&iter,
		gtk_text_buffer_get_insert(buffer)
	);

	gtk_text_buffer_move_mark(buffer, priv->context_mark, &iter);

	return GTK_WIDGET_CLASS(talkatu_view_parent_class)->popup_menu(widget);
}

static void
talkatu_view_emit_send_message_cb(GtkMenuItem *item, gpointer data) {
	TalkatuView *view = TALKATU_VIEW(data);

	g_signal_emit(view, signals[SIG_SEND_MESSAGE], 0);
}

static void
talkatu_view_open_url_cb(GtkMenuItem *item, gpointer data) {
	TalkatuViewOpenLinkData *open_link_data = (TalkatuViewOpenLinkData *)data;

	g_signal_emit(
		open_link_data->view,
		signals[SIG_OPEN_URL],
		0,
		open_link_data->url
	);

	g_slice_free(TalkatuViewOpenLinkData, data);
}

static void
talkatu_view_copy_url_cb(GtkMenuItem *item, gpointer data) {
	GtkClipboard *clipboard = gtk_widget_get_clipboard(
		GTK_WIDGET(item),
		GDK_SELECTION_CLIPBOARD
	);

	gtk_clipboard_set_text(clipboard, (gchar *)data, -1);
}

static void
talkatu_view_populate_popup_cb(GtkTextView *view, GtkWidget *popup) {
	TalkatuViewPrivate *priv = NULL;
	GtkTextBuffer *buffer = NULL;
	GtkTextIter iter;
	GtkWidget *item = NULL;
	gint pos = 0;

	/* if the popup isn't a menu, bail */
	if(!GTK_IS_MENU(popup)) {
		return;
	}

	priv = talkatu_view_get_instance_private(TALKATU_VIEW(view));

	buffer = gtk_text_view_get_buffer(view);

	gtk_text_buffer_get_iter_at_mark(buffer, &iter, priv->context_mark);

	/* add the send message item */
	if(gtk_text_view_get_editable(view)) {
		item = gtk_menu_item_new_with_label(_("Send message"));
		g_signal_connect_after(
			G_OBJECT(item),
			"activate",
			G_CALLBACK(talkatu_view_emit_send_message_cb),
			view
		);
		gtk_menu_shell_insert(GTK_MENU_SHELL(popup), item, pos++);
		gtk_widget_show(item);

		item = gtk_separator_menu_item_new();
		gtk_menu_shell_insert(GTK_MENU_SHELL(popup), item , pos++);
		gtk_widget_show(item);
	}
}

/******************************************************************************
 * GObject Stuff
 *****************************************************************************/
static void
talkatu_view_get_property(GObject *obj, guint prop_id, GValue *value, GParamSpec *pspec) {
	TalkatuView *view = TALKATU_VIEW(obj);

	switch(prop_id) {
		case PROP_SEND_BINDING:
			g_value_set_flags(value, talkatu_view_get_send_binding(view));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
			break;
	}
}

static void
talkatu_view_set_property(GObject *obj, guint prop_id, const GValue *value, GParamSpec *pspec) {
	TalkatuView *view = TALKATU_VIEW(obj);

	switch(prop_id) {
		case PROP_SEND_BINDING:
			talkatu_view_set_send_binding(view, g_value_get_flags(value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
			break;
	}
}

static void
talkatu_view_init(TalkatuView *view) {
	TalkatuViewPrivate *priv = talkatu_view_get_instance_private(view);

	/* set our event mask for the signals we care about */
	gtk_widget_set_events(
		GTK_WIDGET(view),
		GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK
	);

	/* we need to know when the buffer is changed in our parent so we can
	 * update our actions and other stuff.
	 */
	g_signal_connect(
		G_OBJECT(view),
		"notify::buffer",
		G_CALLBACK(talkatu_view_buffer_set_cb),
		NULL
	);

	/* setup GSpell */
	priv->gspell_view = gspell_text_view_get_from_gtk_text_view(GTK_TEXT_VIEW(view));

	/* we need to connect this signal *AFTER* everything to make sure our items
	 * end up in the correct place.
	 */
	g_signal_connect_after(
		G_OBJECT(view),
		"populate-popup",
		G_CALLBACK(talkatu_view_populate_popup_cb),
		NULL
	);
}

static void
talkatu_view_class_init(TalkatuViewClass *klass) {
	GObjectClass *obj_class = G_OBJECT_CLASS(klass);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(klass);
	GtkBindingSet *binding_set = NULL;

	obj_class->get_property = talkatu_view_get_property;
	obj_class->set_property = talkatu_view_set_property;

	widget_class->popup_menu = talkatu_view_popup_menu;

	/* add our default signal handlers */
	klass->format_activate = talkatu_view_format_activate;
	klass->should_send_message = talkatu_view_should_send_message;

	/* add our properties */
	properties[PROP_SEND_BINDING] = g_param_spec_flags(
		"send-binding", "send-binding", "The keybindings that will trigger the send signal",
		TALKATU_TYPE_VIEW_SEND_BINDING,
		TALKATU_VIEW_SEND_BINDING_RETURN | TALKATU_VIEW_SEND_BINDING_KP_ENTER,
		G_PARAM_READWRITE | G_PARAM_CONSTRUCT
	);
	g_object_class_install_properties(obj_class, N_PROPERTIES, properties);

	/* add our signals */

	/**
	 * TalkatuView::format-activate
	 * @talkatutextview: The #TalkatuView instance.
	 * @arg1: The name of the action to activated.
	 * @user_data: User supplied data.
	 *
	 * Emitted by the keybindings to apply a format to the underlying buffer.
	 */
	signals[SIG_FORMAT_ACTIVATE] = g_signal_new(
		"format-activate",
		G_TYPE_FROM_CLASS(klass),
		G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		G_STRUCT_OFFSET(TalkatuViewClass, format_activate),
		NULL,
		NULL,
		NULL,
		G_TYPE_NONE,
		1,
		G_TYPE_STRING
	);

	/**
	 * TalkatuView::should-send-message:
	 * @talkatutextview: The #TalkatuView instance.
	 * @arg1: The #TalkatuViewSendBinding that was entered.
	 * @user_data: User supplied data.
	 *
	 * Emitted when a potential keybinding to send the message is entered to
	 * determine if the message should be sent.
	 */
	signals[SIG_SHOULD_SEND_MESSAGE] = g_signal_new(
		"should-send-message",
		G_TYPE_FROM_CLASS(klass),
		G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		G_STRUCT_OFFSET(TalkatuViewClass, should_send_message),
		NULL,
		NULL,
		NULL,
		G_TYPE_NONE,
		1,
		TALKATU_TYPE_VIEW_SEND_BINDING
	);

	/**
	 * TalkatuView::send-message:
	 * @talkatutextview: The #TalkatuView instance.
	 * @user_data: User supplied data.
	 *
	 * Emitted when a message should be sent.
	 */
	signals[SIG_SEND_MESSAGE] = g_signal_new(
		"send-message",
		G_TYPE_FROM_CLASS(klass),
		G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		G_STRUCT_OFFSET(TalkatuViewClass, send_message),
		NULL,
		NULL,
		NULL,
		G_TYPE_NONE,
		0
	);

	/**
	 * TalkatuView::open-url:
	 * @talkatutextview: The #TalkatuView instances.
	 * @url: The URL to open.
	 * @user_data: User supplied data.
	 *
	 * Emitted when a user clicks on a link to open the url
	 */
	signals[SIG_OPEN_URL] = g_signal_new(
		"open-url",
		G_TYPE_FROM_CLASS(klass),
		G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		G_STRUCT_OFFSET(TalkatuViewClass, open_url),
		NULL,
		NULL,
		NULL,
		G_TYPE_NONE,
		1,
		G_TYPE_STRING
	);

	/* setup key bindings */
	binding_set = gtk_binding_set_by_class(talkatu_view_parent_class);

	/* remove existing keybindings that we're overriding */
	gtk_binding_entry_remove(binding_set, GDK_KEY_slash, GDK_CONTROL_MASK);

	/* add our custom keybindings */
	gtk_binding_entry_add_signal(binding_set, GDK_KEY_b, GDK_CONTROL_MASK, "format-activate", 1, G_TYPE_STRING, TALKATU_ACTION_FORMAT_BOLD);
	gtk_binding_entry_add_signal(binding_set, GDK_KEY_i, GDK_CONTROL_MASK, "format-activate", 1, G_TYPE_STRING, TALKATU_ACTION_FORMAT_ITALIC);
	gtk_binding_entry_add_signal(binding_set, GDK_KEY_u, GDK_CONTROL_MASK, "format-activate", 1, G_TYPE_STRING, TALKATU_ACTION_FORMAT_UNDERLINE);
	gtk_binding_entry_add_signal(binding_set, GDK_KEY_slash, GDK_CONTROL_MASK, "format-activate", 1, G_TYPE_STRING, TALKATU_ACTION_FORMAT_STRIKETHROUGH);
	gtk_binding_entry_add_signal(binding_set, GDK_KEY_plus, GDK_CONTROL_MASK, "format-activate", 1, G_TYPE_STRING, TALKATU_ACTION_FORMAT_GROW);
	gtk_binding_entry_add_signal(binding_set, GDK_KEY_equal, GDK_CONTROL_MASK, "format-activate", 1, G_TYPE_STRING, TALKATU_ACTION_FORMAT_GROW);
	gtk_binding_entry_add_signal(binding_set, GDK_KEY_minus, GDK_CONTROL_MASK, "format-activate", 1, G_TYPE_STRING, TALKATU_ACTION_FORMAT_SHRINK);
	gtk_binding_entry_add_signal(binding_set, GDK_KEY_r, GDK_CONTROL_MASK, "format-activate", 1, G_TYPE_STRING, TALKATU_ACTION_FORMAT_RESET);

	gtk_binding_entry_add_signal(binding_set, GDK_KEY_Return, 0, "should-send-message", 1, TALKATU_TYPE_VIEW_SEND_BINDING, TALKATU_VIEW_SEND_BINDING_RETURN);
	gtk_binding_entry_add_signal(binding_set, GDK_KEY_Return, GDK_SHIFT_MASK, "should-send-message", 1, TALKATU_TYPE_VIEW_SEND_BINDING, TALKATU_VIEW_SEND_BINDING_SHIFT_RETURN);
	gtk_binding_entry_add_signal(binding_set, GDK_KEY_Return, GDK_CONTROL_MASK, "should-send-message", 1, TALKATU_TYPE_VIEW_SEND_BINDING, TALKATU_VIEW_SEND_BINDING_CONTROL_RETURN);
	gtk_binding_entry_add_signal(binding_set, GDK_KEY_KP_Enter, 0, "should-send-message", 1, TALKATU_TYPE_VIEW_SEND_BINDING, TALKATU_VIEW_SEND_BINDING_KP_ENTER);
}

/******************************************************************************
 * Public API
 *****************************************************************************/

/**
 * talkatu_view_new:
 *
 * Creates a new #TalkatuView with a #TalkatuBuffer.
 *
 * Returns: (transfer full): The new #TalkatuView.
 */
GtkWidget *talkatu_view_new(void) {
	return talkatu_view_new_with_buffer(talkatu_buffer_new(NULL));
}

/**
 * talkatu_view_new_with_buffer:
 * @buffer: A #GtkTextBuffer.
 *
 * Creates a new #TalkatuView with @buffer.
 *
 * Returns: (transfer full): The new #TalkatuView.
 */
GtkWidget *talkatu_view_new_with_buffer(GtkTextBuffer *buffer) {
	return GTK_WIDGET(g_object_new(
		TALKATU_TYPE_VIEW,
		"buffer", buffer,
		NULL));
}

/**
 * talkatu_view_set_send_binding:
 * @view: The #TalkatuView instance.
 * @bindings: The #TalkatuViewSendBinding value.
 *
 * Sets the bindings for when the send-message signal should be emitted.
 */
void
talkatu_view_set_send_binding(TalkatuView *view, TalkatuViewSendBinding bindings) {
	TalkatuViewPrivate *priv = NULL;

	g_return_if_fail(TALKATU_IS_VIEW(view));

	priv = talkatu_view_get_instance_private(view);

	priv->send_binding = bindings;

	g_object_notify_by_pspec(G_OBJECT(view), properties[PROP_SEND_BINDING]);
}

/**
 * talkatu_view_get_send_binding:
 * @view: The #TalkatuView instance.
 *
 * Gets the #TalkatuViewSendBinding which determines when send-message
 * signal will be emitted.
 *
 * Returns: The #TalkatuViewSendBinding.
 */
TalkatuViewSendBinding
talkatu_view_get_send_binding(TalkatuView *view) {
	TalkatuViewPrivate *priv = NULL;

	g_return_val_if_fail(TALKATU_IS_VIEW(view), 0);

	priv = talkatu_view_get_instance_private(view);

	return priv->send_binding;
}
